<?php

declare(strict_types=1);

namespace Drupal\cqri\HookHandler;

use Drupal\Core\DependencyInjection\ContainerInjectionInterface;
use Drupal\Core\Entity\EntityTypeManagerInterface;
use Drupal\Core\Image\ImageFactory;
use Drupal\Core\Template\Attribute;
use Drupal\breakpoint\BreakpointInterface;
use Drupal\breakpoint\BreakpointManagerInterface;
use Symfony\Component\DependencyInjection\ContainerInterface;

/**
 * Prepare variables for cqri_item.
 */
class PreprocessCqriItem implements ContainerInjectionInterface {

  public function __construct(
    protected EntityTypeManagerInterface $entityTypeManager,
    protected BreakpointManagerInterface $breakpointManager,
    protected ImageFactory $imageFactory,
  ) {}

  /**
   * {@inheritdoc}
   */
  public static function create(ContainerInterface $container): static {
    return new static(
      $container->get('entity_type.manager'),
      $container->get('breakpoint.manager'),
      $container->get('image.factory'),
    );
  }

  /**
   * Prepare variables for cqri_item.
   *
   * Default template: cqri-item.html.twig.
   *
   * @param array $variables
   *   An associative array containing:
   *   - uri: The URI of the image.
   *   - width: The width of the image (if known).
   *   - height: The height of the image (if known).
   *   - attributes: Associative array of attributes to be placed in the img
   *     tag.
   *   - responsive_image_style_id: The ID of the responsive image style.
   */
  public function preprocess(array &$variables): void {
    \template_preprocess_responsive_image($variables);

    unset($variables['attributes']['srcset'], $variables['attributes']['sizes']);

    $variables['output_image_tag'] = FALSE;

    // Preprocess sources.
    $responsive_image_style = $this->entityTypeManager
      ->getStorage('responsive_image_style')
      ->load($variables['responsive_image_style_id']);
    if ($responsive_image_style == NULL) {
      return;
    }
    // Retrieve all breakpoints and multipliers and reverse order of
    // breakpoints.
    // By default, breakpoints are ordered from the smallest weight to largest:
    // the smallest weight is expected to have the smallest breakpoint width,
    // while the largest weight is expected to have the largest breakpoint
    // width. For responsive images, we need the largest breakpoint widths
    // first, so we need to reverse the order of these breakpoints.
    $breakpoints = \array_reverse($this->breakpointManager->getBreakpointsByGroup($responsive_image_style->getBreakpointGroup()));
    foreach ($responsive_image_style->getKeyedImageStyleMappings() as $breakpoint_id => $multipliers) {
      if (isset($breakpoints[$breakpoint_id])) {
        $variables['sources'] = $this->buildSourceAttributes($variables, $breakpoints[$breakpoint_id], $multipliers);
      }
    }

    // Preprocess Fallback.
    $variables['img_element'] = [
      '#theme' => 'image',
      '#uri' => \_responsive_image_image_style_url($responsive_image_style->getFallbackImageStyle(), $variables['uri']),
      '#attributes' => [],
    ];
    // Get width and height from fallback responsive image style and transfer
    // them to img tag so browser can do aspect ratio calculation and prevent
    // recalculation of layout on image load.
    $dimensions = \responsive_image_get_image_dimensions(
      $responsive_image_style->getFallbackImageStyle(),
      [
        'width' => $variables['width'],
        'height' => $variables['height'],
      ],
      $variables['uri']
    );
    $variables['img_element']['#width'] = $dimensions['width'];
    $variables['img_element']['#height'] = $dimensions['height'];

    if (isset($variables['attributes'])) {
      if (isset($variables['attributes']['alt'])) {
        $variables['img_element']['#alt'] = $variables['attributes']['alt'];
        unset($variables['attributes']['alt']);
      }
      if (isset($variables['attributes']['title'])) {
        $variables['img_element']['#title'] = $variables['attributes']['title'];
        unset($variables['attributes']['title']);
      }
      if (isset($variables['img_element']['#width'])) {
        $variables['attributes']['width'] = $variables['img_element']['#width'];
      }
      if (isset($variables['img_element']['#height'])) {
        $variables['attributes']['height'] = $variables['img_element']['#height'];
      }
      $variables['img_element']['#attributes'] = $variables['attributes'];
    }
  }

  /**
   * Build Source attributes for the "picture" Dom Element.
   *
   * @param array $variables
   *   An array with the following keys:
   *   - responsive_image_style_id:
   *     The \Drupal\responsive_image\Entity\ResponsiveImageStyle ID.
   *   - width: The width of the image (if known).
   *   - height: The height of the image (if known).
   *   - uri: The URI of the image file.
   * @param \Drupal\breakpoint\BreakpointInterface $breakpoint
   *   The breakpoint for this source tag.
   * @param array $multipliers
   *   An array with multipliers as keys and image style mappings as values.
   *
   * @return \Drupal\Core\Template\Attribute[]
   *   An array of attributes for the source tag.
   *
   * @see _responsive_image_build_source_attributes
   */
  protected function buildSourceAttributes(array $variables, BreakpointInterface $breakpoint, array $multipliers): array {
    // @todo $breakpoint is not used. To remove?
    if ((empty($variables['width']) || empty($variables['height']))) {
      $image = $this->imageFactory->get($variables['uri']);
      $width = $image->getWidth();
      $height = $image->getHeight();
    }
    else {
      $width = $variables['width'];
      $height = $variables['height'];
    }
    $extension = \pathinfo($variables['uri'], \PATHINFO_EXTENSION);
    $sizes = [];
    $srcset = [];
    $derivative_mime_types = [];

    // Traverse the multipliers in reverse so the largest image is processed
    // last.
    // The last image's dimensions are used for img.srcset height and width.
    foreach (\array_reverse($multipliers) as $image_style_mapping) {
      if ($image_style_mapping['image_mapping_type'] == 'sizes') {
        // Loop through the image styles for this breakpoint and multiplier.
        foreach ($image_style_mapping['image_mapping']['sizes_image_styles'] as $image_style_name) {
          // Get the dimensions.
          $dimensions = \responsive_image_get_image_dimensions($image_style_name, [
            'width' => $width,
            'height' => $height,
          ], $variables['uri']);

          // Get MIME type.
          $derivative_mime_type = \responsive_image_get_mime_type($image_style_name, $extension);
          $derivative_mime_types[] = $derivative_mime_type;

          // Add the image source with its width descriptor. When a width
          // descriptor is used in a srcset, we can't add a multiplier to
          // it. Because of this, the image styles for all multipliers of
          // this breakpoint should be merged into one srcset and the sizes
          // attribute should be merged as well.
          if ($dimensions['width'] === NULL) {
            throw new \LogicException("Could not determine image width for '{$variables['uri']}' using image style with ID: {$image_style_name}. This image style can not be used for a responsive image style mapping using the 'sizes' attribute.");
          }

          // Use the image width as key so we can sort the array later on.
          // Images within a srcset should be sorted from small to large, since
          // the first matching source will be used.
          $srcset[(int) ($dimensions['width'])] = $dimensions;
          $srcset[(int) ($dimensions['width'])]['srcset'] = \_responsive_image_image_style_url($image_style_name, $variables['uri']);
        }
        $sizes = \array_merge(\explode(', ', $image_style_mapping['image_mapping']['sizes']), $sizes);
      }
    }

    // Sort the srcset from small to large image width or multiplier.
    \krsort($srcset);
    $srcset = \array_values($srcset);

    $sources = [];
    foreach ($srcset as $key => $data) {
      $source_attribute = new Attribute($data);
      $source_attribute->setAttribute('container', $sizes[$key]);

      if (\count(\array_unique($derivative_mime_types)) == 1) {
        $source_attribute->setAttribute('type', $derivative_mime_types[0]);
      }

      $sources[] = $source_attribute;
    }

    return $sources;
  }

}
